using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

// ReSharper disable FieldCanBeMadeReadOnly.Local
// ReSharper disable UnusedMember.Global
// ReSharper disable MemberCanBePrivate.Global
namespace Shared
{
   /// <summary>
   /// Some of the ideas have been taken from the CommandLink Button,
   /// And All of the colours have been copied directly from there.
   /// </summary>
   [ToolboxBitmap(typeof(TabControl))]
   public sealed class RoundedTabControl : TabControl
   {
      #region Designer things

      private Color highlightColor = Color.FromKnownColor(KnownColor.Highlight);

      /// <summary>
      /// Gets and sets the color used for the highlight.
      /// </summary>
      [Category("Highlight")]
      [Description("The color to use for the highlight.")]
      [DefaultValue(typeof(Color), "GradientActiveCaption")]
      public Color HighlightColor
      {
         get { return highlightColor; }
         set
         {
            if (highlightColor != value)
            {
               highlightColor = value;
               Refresh();
            }
         }
      }

      private int highlightWidth = 2;
      /// <summary>
      /// Gets and sets the width of the highlight.
      /// </summary>
      [Category("Highlight")]
      [Description("The width of the border line.")]
      public int HighlightWidth
      {
         get { return highlightWidth; }
         set
         {
            if (value < 1)
               value = 1;
            if (highlightWidth != value)
            {
               highlightWidth = value;
               Refresh();
            }
         }
      }

      private int rounding = 17;
      /// <summary>
      /// Gets and sets the rounding factor controlling how round the corners of the highlight are.
      /// </summary>
      [Category("Highlight")]
      [Description("Controls how round the corners of the highlight are.")]
      public int Rounding
      {
         get { return rounding; }
         set
         {
            if (rounding < 1)
               rounding = 1;
            if (rounding != value)
            {
               rounding = value;
               Refresh();
            }
         }
      }

      private bool useHighlighFill = true;
      /// <summary>
      /// Gets and sets whether the interior background colour of the control will be blended with the highlight
      /// color when the highlight is shown.
      /// </summary>
      [Category("Highlight")]
      [Description("When true the interior background of the control will be blended with the highlight color.")]
      [DefaultValue(true)]
      public bool UseHighlighFill
      {
         get { return useHighlighFill; }
         set
         {
            if (useHighlighFill != value)
            {
               useHighlighFill = value;
               Refresh();
            }
         }
      }

      private byte highlightFillAlphaNormal = 50;

      /// <summary>
      /// Gets and sets the maximum value of the alpha channel used when the control is being drawn normally.
      /// </summary>
      [Category("Highlight")]
      [DefaultValue(50)]
      [Description("Sets the maximum alpha value to use for the graduated normal fill.")]
      public byte HighlightFillAlphaNormal
      {
         get { return highlightFillAlphaNormal; }
         set
         {
            if (highlightFillAlphaNormal != value)
            {
               highlightFillAlphaNormal = value;
               Refresh();
            }
         }
      }


      private byte highlightFillAlpha = 200;

      /// <summary>
      /// Gets and sets the maximum value of the alpha channel used when the control is focus highlighted.
      /// </summary>
      [Category("Highlight")]
      [DefaultValue(200)]
      [Description("Sets the maximum alpha value to use for the graduated highlight fill.")]
      public byte HighlightFillAlpha
      {
         get { return highlightFillAlpha; }
         set
         {
            if (highlightFillAlpha != value)
            {
               highlightFillAlpha = value;
               Refresh();
            }
         }
      }

      private byte highlightFillAlphaMouse = 100;

      /// <summary>
      /// Gets and sets the maximum value of the alpha channel used when the control is mouse highlighted.
      /// </summary>
      [Category("Highlight")]
      [DefaultValue(100)]
      [Description("Sets the maximum alpha value to use for the graduated highlight fill when the the mouse is over the control.")]
      public byte HighlightFillAlphaMouse
      {
         get { return highlightFillAlphaMouse; }
         set
         {
            if (highlightFillAlphaMouse != value)
            {
               highlightFillAlphaMouse = value;
               Refresh();
            }
         }
      }


      #endregion

      public RoundedTabControl()
      {
         SetStyle(ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint | ControlStyles.ResizeRedraw | ControlStyles.OptimizedDoubleBuffer, true);
         showUnderLine = SystemInformation.MenuAccessKeysUnderlined;
         ChangeUICues += btnChangeUICues;
      }

      bool showUnderLine;
      private void btnChangeUICues(object sender, UICuesEventArgs uiCuesEventArgs)
      {
         if (uiCuesEventArgs.ChangeKeyboard)
            showUnderLine = uiCuesEventArgs.ShowKeyboard;
      }

      protected override void OnHandleCreated(EventArgs e)
      {
         base.OnHandleCreated(e);
         if (DesignMode)
            showUnderLine = true;
      }


#region Mouse Over things
      [Browsable(false)]
      private int MouseOverTabIndex
      {
         get
         {
            Point cursor = PointToClient(Cursor.Position);
            int index = -1;
            for (int i = 0; i < TabPages.Count; i++)
            {
               if (GetTabRect(i).Contains(cursor))
               {
                  index = i;
                  break;
               }
            }
            if ((index != -1)
               && !TabPages[index].Enabled
               )
            {
               index = -1;
            }
            return index;
         }
      }

      private int lastMouseOver = -1;
      protected override void OnMouseMove(MouseEventArgs e)
      {
         base.OnMouseMove(e);
         if (HotTrack)
         {
            int thisIndex = MouseOverTabIndex;
            if (lastMouseOver != thisIndex)
            {
               lastMouseOver = thisIndex;
               Invalidate();
            }
         }
      }
      protected override void OnMouseLeave(EventArgs e)
      {
         base.OnMouseLeave(e);
         if (HotTrack)
         {
            lastMouseOver = -1;
            Invalidate();
         }
      }

      #endregion

#region Lets deal with the painting

      protected override void OnPaint(PaintEventArgs e)
      {
         //	We must always paint the entire area of the tab control
         if (e.ClipRectangle.Equals(ClientRectangle))
         {
            for (int index = 0; index < TabPages.Count; index++)
            {
               if ( index != SelectedIndex )
                  DrawTabContents(index, e.Graphics);
            }
            // Make sure the focused tab is drawn topmost
            if (SelectedIndex > -1)
            {
               DrawTabContents(SelectedIndex, e.Graphics);
            }
            if ( TabPages.Count > 0 )
               DrawTabPageBorder(SelectedIndex, FindHighlightColor(Enabled, SelectedIndex, KnownColor.ActiveBorder), e.Graphics);
         }
         else
         {
            //	it is less intensive to just reinvoke the paint with the whole surface available to draw on.
            Invalidate();
         }
      }

      private void DrawTabPageBorder(int index, Color findHighlightColor, Graphics graphics)
      {
         graphics.SmoothingMode = SmoothingMode.HighQuality;
         Rectangle pageBounds = TabPages[index].Bounds;
         if (highlightWidth > 1)
            pageBounds.Inflate(highlightWidth-1, highlightWidth-1);
         using (Pen borderPen = new Pen(findHighlightColor, highlightWidth))
         {
            graphics.DrawRectangle(borderPen, pageBounds);
         }

      }

      /// <summary>
      /// Taken from http://social.msdn.microsoft.com/Forums/fi-FI/winforms/thread/e38246ba-707e-48e8-a5a4-00ac3ace27c5
      /// and then modified
      /// </summary>
      private void DrawTabContents(int index, Graphics graphics)
      {
         Rectangle tabBounds = GetTabRect(index);
         using (Brush b = new SolidBrush(TabPages[index].BackColor))
         {
            graphics.FillRectangle(b, tabBounds);
         }

         bool enabled = TabPages[index].Enabled;
         using (GraphicsPath gp = GetTabBorderPath(tabBounds))
         {
            // Fill the header gradient
            PaintTab(gp, tabBounds, FindHighlightColor(enabled, index, KnownColor.ActiveCaptionText), index == SelectedIndex, graphics);

            // Draw the outline
            using ( Pen p = new Pen(FindHighlightColor(true, SelectedIndex, KnownColor.ActiveBorder),
                               SelectedIndex == index ? highlightWidth : enabled ? 1 : 0.1f))
            {
               graphics.DrawPath(p, gp);
            }
         }

         //Draw the text
         DrawTabText(tabBounds, enabled, index, graphics);
      }

      private Color FindHighlightColor(bool enabled, int index, KnownColor defaultColor)
      {
         Color borderColor;
         if (!enabled)
         {
            borderColor = Color.FromArgb(128, Color.FromKnownColor(KnownColor.GrayText));
         }
         else if (useHighlighFill)
         {
            byte alpha = index == lastMouseOver
                            ? ((index == SelectedIndex) ? byte.MaxValue : highlightFillAlphaMouse)
                            : ((index == SelectedIndex) ? highlightFillAlpha : highlightFillAlphaNormal);
            borderColor = Color.FromArgb(alpha, highlightColor);
         }
         else
         {
            borderColor = Color.FromKnownColor(defaultColor);
         }
         return borderColor;
      }

      private void DrawTabText(Rectangle tabBounds, bool enabled, int index, Graphics graphics)
      {
         graphics.SmoothingMode = SmoothingMode.HighQuality;
         graphics.TextRenderingHint = System.Drawing.Text.TextRenderingHint.ClearTypeGridFit;
         TextFormatFlags tf = TextFormatFlags.WordEllipsis | TextFormatFlags.HorizontalCenter | TextFormatFlags.VerticalCenter;
         if (showUnderLine )
            tf |= TextFormatFlags.HidePrefix;
         Color textColor = enabled ? TabPages[index].ForeColor : Color.FromKnownColor(KnownColor.GrayText);
         TextRenderer.DrawText(graphics, TabPages[index].Text, Font, tabBounds, textColor, Color.Transparent, tf);
      }


      private new Rectangle GetTabRect(int index)
      {
         return (index < 0) ? new Rectangle() : base.GetTabRect(index);
      }

      private GraphicsPath GetTabBorderPath(Rectangle tabBounds)
      {
         // Constant for the rounded rectangle.
         int doubleRadius = rounding * 2;

         GraphicsPath path = new GraphicsPath();
         path.AddArc(tabBounds.X, tabBounds.Y, doubleRadius, doubleRadius, 180, 90);
         path.AddLine(tabBounds.X + rounding, tabBounds.Y, tabBounds.X + tabBounds.Width - rounding, tabBounds.Y);
         path.AddArc(tabBounds.X + tabBounds.Width - doubleRadius, tabBounds.Y, doubleRadius, doubleRadius, 270, 90);
         path.AddLine(tabBounds.X + tabBounds.Width, tabBounds.Y + rounding, tabBounds.X + tabBounds.Width, tabBounds.Y + tabBounds.Height);
         path.AddLine(tabBounds.X + tabBounds.Width, tabBounds.Y + tabBounds.Height, tabBounds.X, tabBounds.Y + tabBounds.Height);
         path.AddLine(tabBounds.X, tabBounds.Y + tabBounds.Height, tabBounds.X, tabBounds.Y + rounding);
         path.CloseFigure();

         return path;
      }


      private void PaintTab(GraphicsPath tabpath, Rectangle tabBounds, Color fill, bool activeTab, Graphics graphics)
      {
         graphics.SmoothingMode = SmoothingMode.HighQuality;
         if (activeTab
            && (highlightWidth > 1)
            )
         {
            tabBounds.Inflate(1 - highlightWidth, 1 - highlightWidth);
         }
         using (Brush b = new LinearGradientBrush(tabBounds, BackColor, fill, activeTab ? 90f : 270f))
         {
            graphics.FillPath(b, tabpath);
         }
      }
      #endregion

   }
}
// ReSharper restore MemberCanBePrivate.Global
// ReSharper restore UnusedMember.Global
// ReSharper restore FieldCanBeMadeReadOnly.Local
